"""
Copyright (C) 2017, 2018, 2019 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------
Use this plugin to view and modify neighbor information.

Usage:
    net add neighbor <interface> (<ipv4>|<ipv6>) <mac>
    net del neighbor <interface> (ipv4|ipv6|all|<ipv4>|<ipv6>)
    net show neighbor [<interface>] [ipv4|ipv6|<ipv4>|<ipv6>] [json]
Options:
"""
import os
import sys
import json
import socket
import logging
import subprocess
from tabulate import tabulate
from os import listdir, readlink
from nclu.NetDaemon import NetDaemon
from nclu.utils import check_output
from nclu.plugins import remove_scratch_files_and_backups

NEIGHBOR_SCRATCHPAD = "/run/nclu/neighbor/neighbor.json"

log = logging.getLogger("netd")

# Parse the receieved commands
def execute_command(net_daemon):
    """
    Given an instance of NetDaemon that has already run an input command through NetworkDocopt,
    perform the appropriate actions, and return the standard 2-tuple.
    """
    assert isinstance(net_daemon, NetDaemon)

    # Convenience variable
    tokens = net_daemon.nclu_parser.args
    assert "neighbor" in tokens, "Why are you here?"

    # Parse Tokens
    af = 'all'
    interface = None
    ip_addr = None
    mac = None
    if '<interface>' in tokens:
        interface = tokens['<interface>']
    if '<ipv4>' in tokens:
        af = 'ipv4'
        ip_addr = tokens['<ipv4>']
    elif '<ipv6>' in tokens:
        af = 'ipv6'
        ip_addr = tokens['<ipv6>']
    elif 'ipv4' in tokens:
        af = 'ipv4'
    elif 'ipv6' in tokens:
        af = 'ipv6'

    # Determine Action
    action = None
    if tokens.get("show") is not None: action = 'show'
    elif tokens.get("add") is not None: action = 'add'
    elif tokens.get("del") is not None: action = 'del'

    # Perform an Action (show or add/del)
    if action == 'show':
        build_json = False
        if tokens.get("json") is not None:
            build_json = True
        return show_neighbors(af=af,interface=interface,ip_addr=ip_addr,build_json=build_json)
    elif action == 'add':
        mac = tokens['<mac>']
        return add_or_del_neighbor_entry(action,mac,interface,ip_addr,af)
    elif action == 'del':
        return add_or_del_neighbor_entry(action,mac,interface,ip_addr,af)

    return (False, {"error": ["Unrecognized 'neighbor' plugin command."]})

def show_neighbors(af=None,interface=None,ip_addr=None,build_json=None,return_datastructure=False):
    """
    net show neighbor [<interface>] [all|ipv4|ipv6|<ipv4>|<ipv6>] [json]
    This function collects neighbor information via 'ip neighbor show' and parses it for display.

    Note: We are screen-scaping here due to the lack of options in 
       3.7.10, `ip -j neighbor show` does not exist in this release
       and nlmanager does not support collection of filtered neighbor
       dumps at this time which lends it to be less efficient if we're
       dumping the full neighbor table for each show operation.
    """
    data = []

    # Build the `ip neighbor` command to collect the data
    address_family = []
    if af == 'ipv4':
        address_family = ['-4']
    elif af == 'ipv6':
        address_family = ['-6']

    interface_chunk = []
    if interface:
        interface_chunk = ['dev',interface]
    neighbor_chunk = []
    if ip_addr:
        neighbor_chunk = [ip_addr]

    show_command = ['/sbin/ip']+address_family+['neighbor','show']+interface_chunk+neighbor_chunk

    # Grab the Show output and handle errors.
    try:
        ip_command_output = check_output(show_command,
                                         stderr=subprocess.STDOUT
                                        ).split('\n')[:-1]
    except subprocess.CalledProcessError as e:
        error_message = ["Could not fetch neighbor information for interface: {}".format(interface),
                         "kernel responds: {}".format(e.output)
                        ]
        for line in error_message: log.error(line)
        return(False,{'error':error_message})

    # Parse the returned data
    # Everyone loves screen-scraping right? Just kidding
    #   We're doing it here again because better primitives are
    #   not universally available here. The output here is at
    #   least very consistent as it does not see much change.
    #   See the "Note" under the show_neighbors function for more.
    for line in ip_command_output:
        log.debug(line)
        neighbor_entry = line.split()
        neighbor_entry_len = len(neighbor_entry)

        # Build our interim data structure, could have used a class here
        #  but there are not quite enough items to warrant the complexity

        # Initializing our neighbor data structure.
        # neighbor is always neighbor_entry[0]
        mac = None
        ifname = None
        if interface: ifname = interface
        # state is always neighbor_entry[-1]

        if neighbor_entry_len > 5:
            # Standard IPv4/IPv6 entry
            # IPv6 entry which is a 'router'
            ifname = neighbor_entry[2]
            mac = neighbor_entry[4]

        elif neighbor_entry_len == 5:
            # Standard IPv4/IPv6 entry with device specified
            # IPv6 entry which is a 'router' with device specified
            mac = neighbor_entry[2]

        elif neighbor_entry_len == 4:
            # Standard IPv4/IPv6 entry with device specified
            # IPv6 entry which is a 'router' with device specified
            if neighbor_entry[-1] == "FAILED":
                ifname = neighbor_entry[2]
            else:
                mac = neighbor_entry[2]

        data.append([neighbor_entry[0],
                     mac,
                     ifname,
                     "IPv6" if ':' in neighbor_entry[0] else "IPv4",
                     neighbor_entry[-1]
                    ] 
                   )

    # Build the outstring
    outstring = ""
    if build_json:
        keyed_output = {'neighbors':{}}
        for entry in data:
            keyed_output['neighbors'][entry[0]]= {'mac' : entry[1],
                                                  'interface' : entry[2],
                                                  'af' : entry[3],
                                                  'state' : entry[4]
                                                 }
        if return_datastructure:
            return keyed_output
        outstring = json.dumps(keyed_output)
    else:
        headers = ['Neighbor', 'MAC', 'Interface', 'AF', 'STATE']
        outstring = tabulate(data, headers=headers)

    return (True, {"output": outstring})

def add_or_del_neighbor_entry(action,mac,interface,ip_addr,af):
    """
    net add neighbor <interface> (<ipv4>|<ipv6) <mac>
    net del neighbor <interface> (ipv4|ipv6|all|<ipv4>|<ipv6>)
    This function creates or removes neighbor entries.
    """

    # Keyed first by interface name and then by IP address, this dict stores
    #    operations which are staged for a future commit.
    to_commit = {} 

    # Example Datastructure
    # to_commit['vlan30'] = [('flush',{'action':'del','mac':None,'af':'ipv4'),
    #                        ('1.1.1.1',{'action':'add','mac':'de:ad:be:ef:00:01','af':'ipv4'),
    #                       ]

    # If the scratchpad already exists, pull it in so we can make sure we 
    #   appropriately stomp out changes in the same commit to the same 
    #   neighbor entry. We pick this up for free by using a dictionary here.
    if os.path.isfile(NEIGHBOR_SCRATCHPAD):
        with open(NEIGHBOR_SCRATCHPAD) as json_file:
            to_commit = json.load(json_file)

    if interface not in to_commit:
        to_commit[interface] = []

    # Handle the change
    if action == 'del' and ip_addr == None:
        # This is the flush action
        to_commit[interface].append(('flush', {'action':action,'af':af}))
    elif action == 'del' and ip_addr:
        # Standard del action
        to_commit[interface].append((ip_addr, {'action':action}))
    elif action == 'add':
        # Standard add action

        # Put MAC in correct format.
        stripped_mac = mac.replace(":", "").replace("-", "").replace(".", "").lower()
        sane_mac = ':'.join(stripped_mac[i:i+2] for i in range(0,len(stripped_mac),2))

        to_commit[interface].append((ip_addr, {'action':action,'mac':sane_mac,'af':af}))

    # Write the Scratchpad
    with open(NEIGHBOR_SCRATCHPAD, 'w') as outfile:
        json.dump(to_commit, outfile)

    return (True,{})

def handle_changes_for_commit(perform_changes=False,to_commit=None):
    """
    Walk the 'to_commit' dictionary and if perform_changes is set to true
    actually perform the work of running the commands. If perform_changes
    is false we output a string of the changes which will be performed at
    the next commit.
    """
    output = ''
    result = True

    if not perform_changes:
        # Only executed in the `net pending` case
        output += '-----------------------------------------------------------\n'
        output += 'NCLU will issue the following commands:\n\n'
        output += '    Note: Neighbor modifications do NOT persist across reboots.\n\n'

    for interface in to_commit:
        for (ip_addr,details) in to_commit[interface]:

            # Build the `ip neighbor` command
            ip_command = None
            if ip_addr == 'flush':
                af_chunk = []
                if details['af'] == 'ipv4': af_chunk = ['-4']
                elif details['af'] == 'ipv6': af_chunk = ['-6']
                ip_command = ['/sbin/ip']+af_chunk+['neighbor',
                                                    'flush',
                                                    'dev',
                                                    interface
                                                   ]
            elif details['action'] == 'add':
                ip_command = ['/sbin/ip',
                              'neighbor',
                              'add',
                              ip_addr,
                              'lladdr',
                              details['mac'],
                              'nud',
                              'reachable',
                              'dev',
                              interface
                             ]
            elif details['action'] == 'del':
                ip_command = ['/sbin/ip',
                              'neighbor',
                              'del',
                              ip_addr,
                              'dev',
                              interface
                             ]

            # What command is to be run?
            output += ' '.join(ip_command) + '\n'

            if perform_changes:
                # Issue the `ip neighbor` command, grab the output and handle errors.
                try:
                    ip_command_output = check_output(ip_command,
                                                     stderr=subprocess.STDOUT
                                                    ).split('\n')[:-1]
                    output += "   Success.\n\n"
                except subprocess.CalledProcessError as e:
                    # Handle "file exists" errors for idempotency
                    if details['action'] == 'add' and "file exists" in e.output.lower():
                        # Does this same entry already exist? Or is it just a similar one?
                        # Let's check...
                        keyed_output = show_neighbors(af=details['af'],interface=interface,ip_addr=ip_addr,build_json=True,return_datastructure=True)
                        if details['mac'] != keyed_output['neighbors'][ip_addr]['mac']:
                            # Same entry was not already present
                            output += " ERROR: Failed trying to add this entry ^\n"
                            output += "        There is already a different MAC {}\n".format(keyed_output['neighbors'][ip_addr]['mac'])
                            output += "        associated with this IP address.\n"
                            output += "    Try removing the existing entry first.\n"
                            output += "       net del neighbor {} {}\n\n".format(interface,ip_addr)
                            result = False
                        else:
                            # Same entry WAS already present
                            output += "   Success. (The entry was already there)\n\nn"
                    elif details['action'] == 'del' and "no such file or directory" in e.output.lower():
                        # The entry cannot be removed, it was already gone.
                        output += "   Success. (The entry was already gone)\n\n"
                    else:
                        output += " ERROR: Failed trying to apply this entry ^\n"
                        output += "        --> {}\n".format(e.output)
                        result = False

    if not perform_changes:
        # Only executed in the `net pending` case
        output += '-----------------------------------------------------------\n'

    return (result,output)

def commit_pending(verbose, net_del_all):
    """
    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"; this plug-in doesn't use it
    """

    if not os.path.isfile(NEIGHBOR_SCRATCHPAD):
        # There are no changes to commit.
        return (True, {})

    # Load-in the to_commit dictionary
    to_commit = None
    with open(NEIGHBOR_SCRATCHPAD) as json_file:
        to_commit = json.load(json_file)

    # Apply the changes and grab the output.
    result, output = handle_changes_for_commit(perform_changes=True,to_commit=to_commit)

    if not result:
        # Commit failed.
        # Perhaps we should back out the changes in the future but for now we're going to
        #   leave them all in the buffer.
        log.error(output)
        return (result,{'output':output})
    else:
        # Commit was successful
        remove_scratch_files_and_backups([NEIGHBOR_SCRATCHPAD])

    return (result,{'output':''})

def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """
    if not os.path.isfile(NEIGHBOR_SCRATCHPAD):
        # There are no changes pending
        return ""

    # Load-in the to_commit dictionary
    to_commit = None
    with open(NEIGHBOR_SCRATCHPAD) as json_file:
        to_commit = json.load(json_file)

    # Grab the string of pending output, chuck the 'result' here, irrelevant.
    result, output = handle_changes_for_commit(perform_changes=False,to_commit=to_commit)

    return output

def abort_pending():
    remove_scratch_files_and_backups([NEIGHBOR_SCRATCHPAD])



#################################################
# NCLU Plugin Boiler Plate
#    Not used specifically by this plugin, but 
#    required to be a valid plugin.
#################################################

def del_all():
    """
    net del all

    """
    pass

def show_config_commands(ifupdown2_wrapper):
    """
    net show configuration commands
    """
    return ([], [])

def show_config_summary(summary):
    """ net show configuration """
    pass

def get_managed_files():
    """ Return a list of managed files. """

    return []

def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """
    return ''

def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """
    return []
