"""
Copyright (C) 2017, 2018, 2019 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------
Use this module to specify Port Security parameters.


Usage:
    net show system [json]
    net show system leds [json]
    net show system ztp [json]
    net show system ztp script
    net show system sensors [json]
    net show system sensors detail [json]
    net show system asic [json]
Options:
    leds    : Show the current state of platform LEDs on the outside of the switch
    ztp     : Show details of Zero Touch Provisioning
    script  : Show the contents of the last ZTP script to run
    sensors : Show current status of Environmental Sensors
    asic    : Show current status of ASIC utilization data
"""
import os
import re
import sys
import json
import math
import logging
import subprocess
from os import listdir, readlink
from datetime import datetime, timedelta
from nclu.NetDaemon import NetDaemon
from nclu.plugins import (
    get_config_files_helper,
    run_commands,
)


log = logging.getLogger("netd")


# Parse the receieved commands
def execute_command(net_daemon):
    """
    Given an instance of NetDaemon that has already run an input command through NetworkDocopt,
    perform the appropriate actions, and return the standard 2-tuple.
    """

    assert isinstance(net_daemon, NetDaemon)

    # Convenience variable
    tokens = net_daemon.nclu_parser.args
    assert "system" in tokens, "Why are you here?"

    if tokens.get("show") is not None:
        if tokens.get("sensors") is not None:
            return system_sensors(tokens)
        elif tokens.get("leds") is not None:
            return system_leds(tokens)
        elif tokens.get("ztp") is not None:
            return system_ztp(tokens)
        elif tokens.get("asic") is not None:
            return system_asic(tokens)
        else:
            # Generic "System" command
            return net_show_system(tokens)

    return (False, {"error": ["Unrecognized 'system' plugin command."]})

def system_leds(tokens):
    """
    net show system leds
    This function catches the output from `/usr/sbin/ledmgrd -dj` and formats it for display."
    """
    led_command = ['/usr/sbin/ledmgrd', '-d']

    if tokens.get("json") is not None:
        # Return a JSON conversion of the data.
        led_command = ['/usr/sbin/ledmgrd', '-j']

    try:
        led_output = subprocess.check_output(led_command)
    except:
        outstring = "Could not collect output from command: {}".format(led_command)
        log.warning(outstring)
        return (True, {"output": outstring})

    return (True, {"output": led_output})

def system_ztp(tokens):
    """
    net show system ztp [json]
    net show system ztp script
    This function catches the output from `/usr/sbin/ztp -s` and formats it for display."
    """

    # ZTP Script fetching Path
    if tokens.get("script") is not None:
        try:
            ztp_script_files = subprocess.check_output(["ls -t /var/lib/cumulus/ztp/ztp_script*"],shell=True).split()
        except:
            return (True, {"output": "No ZTP script has been run yet."})

        try:
            with open(ztp_script_files[0],'r') as ztp_script_fd:
                outstring =  "Last ZTP Script stored at: {}\n".format(ztp_script_files[0])
                outstring += "Contents of Last ZTP script:\n"
                for line in ztp_script_fd.readlines():
                    outstring += line
                return (True, {"output": outstring})
        except:
            outstring = "Could not open ZTP script file: {}".format(ztp_script_files(0))
            log.warning(outstring)
            return (False, {"output": outstring})

    # ZTP Status Path
    ztp_command = ['/usr/sbin/ztp', '-s']

    if tokens.get("json") is not None:
        # Return a JSON conversion of the data.
        ztp_command = ['/usr/sbin/ztp', '-j']

    try:
        ztp_output = subprocess.check_output(ztp_command)
    except:
        outstring = "Could not collect output from command: {}".format(ztp_command)
        log.warning(outstring)
        return (True, {"output": outstring})

    return (True, {"output": ztp_output})

def system_sensors(tokens):
    """
    net show system sensors [json]
    net show system sensors detail [json]
    This function catches the output from `/usr/sbin/smonctl` and formats it for display."
    """

    # Default Command shows summary view
    sensors_command = ['/usr/sbin/smonctl']

    if tokens.get("json") is not None:
        # Return a JSON conversion of the data.
        sensors_command = ['/usr/sbin/smonctl', '-j']

    elif tokens.get("detail") is not None:
        # Return a more detailed view of sensor data.
        sensors_command = ['/usr/sbin/smonctl', '-v']

    try:
        sensors_output = subprocess.check_output(sensors_command)
    except subprocess.CalledProcessError as e:
        # smonctl is throwing CalledProcessError with 255 returncode
        # when *any* of the sensors are absent or bad.  We want to
        # show that output so eat the excception, grab the output
        # and show it.
        if e.returncode is 255:
            sensors_output = e.output
        else:
            outstring = "Could not collect output from command: {}".format(sensors_command)
            log.warning(outstring)
            return (True, {"output": outstring})
    except:
        outstring = "Could not collect output from command: {}".format(sensors_command)
        log.warning(outstring)
        return (True, {"output": outstring})


    return (True, {"output": sensors_output})

def system_asic(tokens):
    """
    net show system asic [json]
    This function catches the output from `/usr/cumulus/bin/cl-resource-query` and formats it for display."
    """

    from cumulus.platforms import probe
    platform_object = probe()
    asic_vendor = None
    if platform_object and platform_object.switch and platform_object.switch.chip:
        asic_vendor = platform_object.switch.chip.sw_base
    if asic_vendor == "vx":
        return (True, {"output": "Vx platform does not have a physical ASIC to query."})


    # Default Command shows summary view
    asic_command = ['/usr/cumulus/bin/cl-resource-query']

    if tokens.get("json") is not None:
        # Return a JSON conversion of the data.
        asic_command = ['/usr/cumulus/bin/cl-resource-query', '-j']

    try:
        asic_output = subprocess.check_output(asic_command)
    except:
        outstring = "Could not collect output from command: {}".format(asic_command)
        log.warning(outstring)
        return (False, {"output": outstring})

    return (True, {"output": asic_output})

def get_disk_size():
    """
    Return the disk size as a string.

    If the disk size cannot be determined, returns None.
    """

    try:
        block_devices = subprocess.check_output('lsblk -r'.split())

    except subprocess.CalledProcessError:
        log.warning("Cannot get the block devices.  The disk size cannot be determined.")
        return None

    disk_size = re.search(r"""^sda\s+(?:\S+\s+){2}(?P<size>\S+)""", block_devices, re.MULTILINE)

    return disk_size.group("size") if (disk_size is not None) else None

def get_readable_memory_size(kB_size):
    """ Return a human readable string for the input size (in kB) """

    p = int(math.floor(math.log(kB_size, 2)/10))

    return "{0:.0f}{1}".format(kB_size/math.pow(1024, p), ['KB', 'MB', 'GB', 'TB'][p])

def net_show_system(tokens):
    """
    net show system
    This is the function for the primary use case of showing the main system details.
    """
    data = {}

    # Get the operating system name and version.
    try:
        with open("/etc/lsb-release") as f:
            lsb_release = f.read()
        data['lsb-release'] = lsb_release
    except OSError:
        data['lsb-release'] = False
        log.warning("Could not determine lsb-release.")

    if data['lsb-release']:

        os_version = re.search(
            r"""^\s*DISTRIB_RELEASE\s*=\s*(.+)""",
            data['lsb-release'], re.IGNORECASE | re.MULTILINE)
        if os_version is not None:
            os_version = os_version.group(1)

        os_description = re.search(
            r'''^\s*DISTRIB_DESCRIPTION\s*=\s*"(.+)"''',
            lsb_release, re.IGNORECASE | re.MULTILINE)
        if os_description is not None:
            os_description = os_description.group(1)

        data['os-description'] = os_description
        data['os-version'] = os_version
        data['build'] = os_description

    # Get the platform information (vendor and model).
    try:
        platform_detect = subprocess.check_output('/usr/bin/platform-detect').strip()
        # platform_detect looks like "dell,s6000_s1220".
        vendor, second = platform_detect.split(',')
        model = second.split('_')[0]
        data['platform'] = {
            'detected': True,
            'model': model.upper(),
            'vendor': vendor.title()
        }
    except subprocess.CalledProcessError:
        platform_detect = None
        data['platform'] = {
            'detected': False,
        }

    # Get the chipset, port configuration, and CPU, if available.  This
    # information is stored as JSON in files, but the desired information might
    # not be in the file one would expect from the file names.  So aggregate
    # all the available information in a dictionary first.

    if platform_detect is not None:
        # search from both  new prefered location and legacy location
        json_dirs = (
            "/usr/share/cumulus-platform/common/platform-info/",
            "/usr/share/cumulus/platform-info/"
        )

        aggregate_platform_dict = {}

        for directory in json_dirs:
            if not os.path.isdir(directory):
                continue

            for filename in listdir(directory):

                json_file_path = "{0}{1}".format(directory, filename)
                assert os.path.isfile(json_file_path)

                try:
                    with open(json_file_path) as f:
                        platform_info = f.read()
                except OSError:
                    log.warning("Could not open {0}.".format(json_file_path))
                    continue

                try:
                    platform_info = json.loads(platform_info)
                except ValueError:
                    log.warning("Could not parse {0}.".format(json_file_path))
                    continue

                if not isinstance(platform_info, dict):
                    log.warning("{0} contains JSON, but it is not a dictionary.".format(json_file_path))
                    continue

                for key, value in platform_info.iteritems():
                    # The value is a list containing one dictionary.
                    aggregate_platform_dict[key] = value

        if platform_detect not in aggregate_platform_dict:
            # The platform might be a chassis, which is not currently supported properly in platformdb.
            # (Updates are scheduled for CL 4.0.)

            if platform_detect.startswith("cel,bigstone_"):
                platform_detect = "facebook,backpack"
            elif platform_detect.startswith("accton,omp800_"):
                platform_detect = "accton,OMP-800"

        if platform_detect in aggregate_platform_dict:
            # Found relevant platform information above.  Display it here.  Substitute the empty string for None.

            try:
                data['platform']['info'] = aggregate_platform_dict[platform_detect][0]  # noqa
            except KeyError:
                log.warning("Could not detect platform information for \"{0}\".  Raised KeyError.".format(platform_detect))

        else:
            log.warning("Could not detect platform information for \"{0}\".".format(platform_detect))

    # Get the system uptime.
    if os.path.isfile("/proc/uptime"):
        with open("/proc/uptime") as f:
            # Looks like "182794.74 653446.10", where the first number is the uptime in seconds.
            uptime = str(timedelta(seconds=float(f.read().split()[0])))
        data['uptime'] = uptime
    else:
        data['uptime'] = None
        log.warning("Could not determine uptime.")

    # Get the hostname
    try:
        data['hostname'] = subprocess.check_output(['/bin/hostname']).rstrip()
    except subprocess.CalledProcessError:
        data['hostname'] = None

    # Get the total memory size.
    with open('/proc/meminfo') as f:
        meminfo = f.read()
    matched = re.search(r"""^MemTotal:\s+(\d+)""", meminfo)
    if matched is not None:
        data['memory'] = int(matched.group(1))
    else:
        data['memory'] = None
        log.warning("Could not determine total memory size.")

    # Get the disk size.
    data['disk'] = get_disk_size()

    # Get the raw eeprom data (well, formatted by decode-syseeprom).
    try:
        data['eeprom'] = json.loads(subprocess.check_output('/usr/cumulus/bin/decode-syseeprom --json'.split()))
    except (subprocess.CalledProcessError, ValueError) as e:
        data['eeprom'] = None
        log.exception(e)

    if tokens.get("json") is not None:
        # Return a JSON conversion of the data.
        try:
            return (True, {"output": json.dumps(data)})
        except ValueError as e:
            log.exception(e)
            return (True, {"output": str(e)})

    # Otherwise, output the data in a reasonably organized fashion.
    output = []

    # Hostname
    if data['hostname'] is not None:
        output.append("{0:.<17} {1}".format(
            "Hostname",
            data['hostname'].split('.')[0]))

    # Build
    if data['os-description'] is not None:
        output.append("{0:.<17} {1}".format(
            "Build",
            data['os-description']))

    # Uptime
    if data['uptime'] is not None:
        output.append("{0:.<17} {1}".format(
            "Uptime",
            data['uptime']))

    # seperator
    output.append(" ")

    # Model
    if data['platform']['detected'] is not None:
        output.append("{0:.<17} {1} {2}".format(
            "Model",
            data['platform']['vendor'].title(),
            data['platform']['model'].upper()))

    # CPU
    if 'info' in data['platform']:
        info = data['platform']['info']
        cpu = info['cpu']

        if cpu['arch'] is not None:
            output.append("{0:.<17} {1} {2} {3} {4} {5}".format(
                "CPU",
                cpu['arch'],
                cpu['vendor'],
                cpu['model'],
                cpu['model_id'],
                cpu['speed']))

    # Memory
    if data['memory'] is not None:
        output.append("{0:.<17} {1}".format(
            "Memory",
            get_readable_memory_size(data['memory'])))

    # Disk
    if data['disk'] is not None:
        output.append("{0:.<17} {1}B".format(
            "Disk",
            data['disk']))

    # ASIC and port configuration
    if 'info' in data['platform']:
        info = data['platform']['info']
        soc = info['soc']

        # ASIC
        if soc['vendor'] is not None:
            output.append("{0:.<17} {1} {2} {3}".format(
                "ASIC",
                soc['vendor'],
                soc['model'],
                soc['model_id']))

        # Port configuration
        if info['ports']['layout']:
            output.append("{0:.<17} {1}".format(
                "Ports",
                info['ports']['layout']))

    # EEPROM info.  EEPROMs can vary widely in content (how they are or are not programmed)
    # so display only the TLVs that are present and silently skip over
    # missing pieces.
    if data['eeprom'].get('validTlvInfoHeader') is not None:
        tlvs = data['eeprom'].get('tlv')

        # Display these fields from the EEPROM if they are present.
        display = (
            'Base MAC Address',
            'Manufacturer',
            'ONIE Version',
            'Part Number',
            'Platform Name',
            'Product Name',
            'Serial Number',
            'Service Tag',
            'Vendor Name'
        )

        # Display the items from the display list that are found in the EEPROM
        for item, value in {k:tlvs[k].get('value') for k in display if k in tlvs.keys()}.iteritems():
            if value is not None:
                output.append("{0:.<17} {1}".format(item, value))

    output = '\n'.join(output)

    return (True, {"output": output})

#################################################
# NCLU Plugin Boiler Plate
#    Not used specifically by this plugin, but
#    required to be a valid plugin.
#################################################

def show_config_commands(ifupdown2_wrapper):
    """
    net show configuration commands
    """
    return ([], [])


def show_config_summary(summary):
    """ net show configuration """
    pass

def commit_pending(verbose, net_del_all):
    """
    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"; this plug-in doesn't use it
    """
    return (True,{})


def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """

    return ""

def abort_pending():
    pass

def get_managed_files():
    """ Return a list of managed files. """

    return []


def del_all():
    """
    (Called by NCLU framework code.)
    """
    pass


def get_config_files():
    """
    Get the contents of all existing, non-empty files.

    (Called by NCLU framework code.)

    Returns:
        A string with the content of all existing, non-empty configuration
        files managed with this plug-in.
    """
    return get_config_files_helper(get_managed_files())


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a
    rollback.  Some plug-ins use this in commit_pending as well.

    (Called by NCLU framework code.)

    Args:
        filename: String filename of a file that's been updated.

    Returns:
        A list of command tuples.
    """
    return []
